/*
 * Licensed Materials - Property of IBM
 * 5748-XX8
 * (c) Copyright International Business Machines Corp. 2001
 * All Rights Reserved.
 * US Government Users Restricted Rights - Use, duplicaion or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 */

/* C Includes */

#include <stdio.h>
#include <string.h>

/* aznAPI Includes */

#include <ogauthzn.h>
#include <aznutils.h>
#include <pdbaclmsg.h>
#include <azn_admin_svc_protos.h>

/* Defines */

#define MAXPOBJNAMELENGTH 1024
#define ADMIN_PLUGIN_VERSION     "AZN Administration Service Demo Plugin v1.0"

/* Make sure this file can be compiled using either the C or C++ compiler */
#ifdef __cplusplus
extern "C" {
#endif

/**************************************************************************
 * Interface function definitions.
 **************************************************************************/

/*
 * FUNCTION NAME
 *      azn_svc_initialize
 *
 * DESCRIPTION
 *      init the AZN Administration service
 *
 * ARGUMENTS
 *      [in] argc       The count of arguments to the service.
 *
 *      [in] argv       The array of argument strings.
 *
 *      [in] svcInit    List of initialization attributes for the service.
 *
 *      [out] svcInfo   attr list ptr for attributes returned by the
 *                      service.
 *
 * RETURN VALUE
 *      AZN_S_COMPLETE on success, error code on failure
 */

    AZN_DECLSPEC
    azn_status_t
    AZN_CALLTYPE
    azn_svc_initialize(
                          const int                   argc,           /* in */
                          const char                  **argv,         /* in */
                          const azn_attrlist_h_t      svcInit,        /* in */
                          azn_attrlist_h_t            *svcInfo        /* out */
                          )
    {
        azn_status_t        st;
        azn_boolean_t       freeAttrlist = FALSE;
        int                 i = 0;

        /* svcInfo must not be NULL */
        if (svcInfo == NULL)
            return(azn_util_errcode(AZN_S_SVC_ADMIN_INVALID_SVCINFO_HDL, 0));

        /* ensure argc is valid */
        if (argc < 0 || (argc == 0 && argv != NULL))
            return(azn_util_errcode(AZN_S_SVC_ADMIN_INVALID_ARG_COUNT, 0));

        if (argc > 0 && argv == NULL)
            return(azn_util_errcode(AZN_S_SVC_ADMIN_INVALID_ARG_ARRAY, 0));


        /* print the arguments */
        printf("%s -- azn_svc_initialize() - Parameters --->\n",
               ADMIN_PLUGIN_VERSION);
        for (i=0; i < argc; i++)
            printf("param[%i] = %s\n", i, argv[i]);

        /* return the service version to the dispatcher */
        if (*svcInfo == AZN_C_INVALID_HANDLE) {
            azn_attrlist_create(svcInfo);
            freeAttrlist = TRUE;
        }

        st = azn_attrlist_add_entry(*svcInfo,
                                    azn_svc_version,
                                    ADMIN_PLUGIN_VERSION);

        if (st != AZN_S_COMPLETE) {
            if (freeAttrlist)
                azn_attrlist_delete(svcInfo);
            return(st);
        }

        return(azn_util_errcode(AZN_S_COMPLETE, 0));
    }


/*
 * FUNCTION NAME
 *      azn_svc_shutdown
 *
 * DESCRIPTION
 *      Shutdown the AZN Administration Service.
 *      The initialization parameters are
 *      pass in again on shutdown but are ignored. No version info
 *      is returned.
 *
 * ARGUMENTS
 *      [in] argc       The count of arguments to the service.
 *
 *      [in] argv       The array of argument strings.
 *
 *      [in] svcInit    List of initialization attributes for the service.
 *
 *      [out] svcInfo   attr list ptr for attributes returned by the
 *                      service.
 *
 * RETURN VALUE
 *      AZN_S_COMPLETE on success, error code on failure
 */

    AZN_DECLSPEC
    azn_status_t
    AZN_CALLTYPE
    azn_svc_shutdown(
                        const int                   argc,     /* in */
                        const char                  **argv,   /* in */
                        const azn_attrlist_h_t      svcInit,  /* in */
                        azn_attrlist_h_t      	    *svcInfo  /* out */
                        )
    {
        if (svcInfo != NULL)
            *svcInfo = AZN_C_INVALID_HANDLE;

        return(azn_util_errcode(AZN_S_COMPLETE, 0));
    }

/*
 * FUNCTION NAME
 *      azn_admin_get_object
 *
 * DESCRIPTION
 *      Implementation of the azn_admin_get_object API
 *
 * INPUT PARAMETERS
 *
 * creds Credentials of the identity requesting the object.
 *
 * locale Locale of the caller at the client machine.
 *
 * path  Protected object path.
 *
 * indata  Free form attribute list that can be used for agreed upon
 *         communication between the service implementation and a custom user
 *         interface.
 *
 * OUTPUT PARAMETERS
 *
 * outdata Free form attribute list.
 *
 *         The service inplementation needs to pass the protected object
 *         information in the azn_admin_svc_pobj attribute (using
 *         the azn_attrlist_add_entry_pobj() API).
 *
 *         The service implementation needs to pass the result strings
 *         information in the azn_admin_svc_results attribute (using the
 *         azn_attrlist_add_entry() API). Typically, these strings would
 *         include error, warning, and information messages.
 *
 *         The service implementation and the custom user interface can
 *         exchange information of  their choice by setting other attributes
 *         (using the azn_attrlist_add_entry_* APIs).
 *
 * RETURN VALUE
 *      AZN_S_COMPLETE on success, error code on failure
 */

    AZN_DECLSPEC
    azn_status_t
    AZN_CALLTYPE
    azn_admin_get_object(
                        azn_creds_h_t       creds,
                        azn_string_t        locale,
                        azn_string_t        path,
                        azn_attrlist_h_t    indata,
                        azn_attrlist_h_t    outdata)
    {
        azn_pobj_desc pobj1;
        char          pobjname[MAXPOBJNAMELENGTH];

        printf("%s -- azn_admin_get_object() - ENTRY\n", ADMIN_PLUGIN_VERSION);

        /*
         * Code needs to be added here to check if the object name is one the
         * plugin knows about
        */

        /* set up the pobj */
        pobjname[0] = '\0';
        strcat(pobjname, path);
        pobj1.name =  (char *)&(pobjname[0]);
        pobj1.type = 5;
        pobj1.description = "Object description";
        pobj1.is_policy_attachable = 1;


        /* this is for ivadmin_protobj_exists() test. */

        if( strcmp(path,"/aznadminsvc/java_adminsvc_object/junk") == 0){
                return(azn_util_errcode(AZN_S_FAILURE, ivacl_s_pobj_not_found));
        }

    /*
     * Code needs to be added below to check the return code for all
     * the aznAPI calls
     */

        azn_attrlist_add_entry_pobj(outdata, azn_admin_svc_pobj, &pobj1);
        azn_attrlist_add_entry(outdata, azn_admin_svc_results, "result1");
        azn_attrlist_add_entry(outdata, azn_admin_svc_results, "result2");
        azn_attrlist_add_entry(outdata, azn_admin_svc_results, "result3");

        printf("%s -- azn_admin_get_object() - EXIT\n", ADMIN_PLUGIN_VERSION);

        return(azn_util_errcode(AZN_S_COMPLETE, 0));
    }

/*
 * FUNCTION NAME
 *      azn_admin_get_objectlist
 *
 * DESCRIPTION
 *      Implementation of the azn_admin_get_objectlist API
 *
 * INPUT PARAMETERS
 *
 * creds Credentials of the identity requesting the object.
 *
 * locale Locale of the caller at the client machine.
 *
 * path  Protected object path.
 *
 * indata  Free form attribute list that can be used for agreed upon
 *         communication between the service implementation and a custom user
 *         interface.
 *
 * OUTPUT PARAMETERS
 *
 * outdata Free form attribute list.
 *
 *         The service inplementation needs to pass the protected object
 *         information in the azn_admin_svc_pobj attribute (using
 *         the azn_attrlist_add_entry_pobj() API).
 *         Each protected object information will be a value of this attribute.
 *
 *         The service implementation needs to pass the result strings
 *         information in the azn_admin_svc_results attribute (using the
 *         azn_attrlist_add_entry() API). Typically, these strings would
 *         include error, warning, and information messages.
 *
 *         The service implementation and the custom user interface can
 *         exchange information of  their choice by setting other attributes
 *         (using the azn_attrlist_add_entry_* APIs).
 *
 * RETURN VALUE
 *      AZN_S_COMPLETE on success, error code on failure
 */


    AZN_DECLSPEC
    azn_status_t
    AZN_CALLTYPE
    azn_admin_get_objectlist(
                            azn_creds_h_t       creds,
                            azn_string_t        locale,
                            azn_string_t        path,
                            azn_attrlist_h_t    indata,
                            azn_attrlist_h_t    outdata)
    {
        azn_pobj_desc pobj1;
        azn_pobj_desc pobj2;
        azn_pobj_desc pobj3;

        char pobjname[MAXPOBJNAMELENGTH];

        printf("%s -- azn_admin_get_objectlist() - ENTRY\n", ADMIN_PLUGIN_VERSION);

        /* set up the pobjs; add them to outdata */

        /*
         * Code needs to be added below to check the return code for all
         * the aznAPI calls
         */

        pobjname[0] = '\0';
        strcat(pobjname, path);
        strcat(pobjname, "/testpobj1");
        pobj1.name =  (char *)&(pobjname[0]);
        pobj1.type = 5;
        pobj1.description = "testpobj1 description";
        pobj1.is_policy_attachable = 1;

        azn_attrlist_add_entry_pobj(outdata, azn_admin_svc_pobj, &pobj1);

        pobjname[0] = '\0';
        strcat(pobjname, path);
        strcat(pobjname, "/testpobj2");
        pobj2.name =  (char *)&(pobjname[0]);
        pobj2.type = 5;
        pobj2.description = "testpobj2 description";
        pobj2.is_policy_attachable = 1;

        azn_attrlist_add_entry_pobj(outdata, azn_admin_svc_pobj, &pobj2);

        pobjname[0] = '\0';
        strcat(pobjname, path);
        strcat(pobjname, "/testpobj3");
        pobj3.name =  (char *)&(pobjname[0]);
        pobj3.type = 5;
        pobj3.description = "testpobj3 description";
        pobj3.is_policy_attachable = 1;

        azn_attrlist_add_entry_pobj(outdata, azn_admin_svc_pobj, &pobj3);

        /* add the results strings  to outdata */

        azn_attrlist_add_entry(outdata, azn_admin_svc_results, "result1");
        azn_attrlist_add_entry(outdata, azn_admin_svc_results, "result2");
        azn_attrlist_add_entry(outdata, azn_admin_svc_results, "result3");

        printf("%s -- azn_admin_get_objectlist() - EXIT\n", ADMIN_PLUGIN_VERSION);

        return(azn_util_errcode(AZN_S_COMPLETE, 0));
    }

/*
 * FUNCTION NAME
 *      azn_admin_get_performtask
 *
 * DESCRIPTION
 *      Implementation of the azn_admin_perform_task API
 *
 * INPUT PARAMETERS
 *
 * creds Credentials of the identity requesting the object.
 *
 * locale Locale of the caller at the client machine.
 *
 * command Command string that identifies the task to be performed.
 *
 * indata  Free form attribute list that can be used for agreed upon
 *         communication between the service implementation and a custom user
 *         interface.
 *
 * OUTPUT PARAMETERS
 *
 * outdata Free form attribute list.
 *
 *         The service implementation needs to pass the result strings
 *         information in the azn_admin_svc_results attribute (using the
 *         azn_attrlist_add_entry() API). Typically, these strings would
 *         include error, warning, and information messages.
 *
 *         The service implementation and the custom user interface can
 *         exchange information of  their choice by setting other attributes
 *         (using the azn_attrlist_add_entry_* APIs).
 *
 * RETURN VALUE
 *      AZN_S_COMPLETE on success, error code on failure
 */

    AZN_DECLSPEC
    azn_status_t
    AZN_CALLTYPE
    azn_admin_perform_task(
                          azn_creds_h_t       creds,
                          azn_string_t        locale,
                          azn_string_t        command,
                          azn_attrlist_h_t    indata,
                          azn_attrlist_h_t    outdata)
    {
        printf("%s -- azn_admin_perform_task() - ENTRY\n", ADMIN_PLUGIN_VERSION);

        printf("locale: %s\n", locale);
        printf("command: %s\n", command);

        /*
         * Code needs to be added below to check the return code for all
         * the aznAPI calls
         */

        azn_attrlist_add_entry(outdata, azn_admin_svc_results, "result1");
        azn_attrlist_add_entry(outdata, azn_admin_svc_results, "result2");
        azn_attrlist_add_entry(outdata, azn_admin_svc_results, "result3");

        printf("%s -- azn_admin_perform_task() - EXIT\n", ADMIN_PLUGIN_VERSION);

        return(azn_util_errcode(AZN_S_COMPLETE, 0));
    }

/*
 * FUNCTION NAME
 *      azn_admin_get_tasklist
 *
 * DESCRIPTION
 *      Implementation of the azn_admin_get_tasklist API
 *
 * INPUT PARAMETERS
 *
 * creds Credentials of the identity requesting the object.
 *
 * locale Locale of the caller at the client machine.
 *
 * commands[] NULL terminated list of administration commands supported
 *            by this service implementation.
 *
 * indata  Free form attribute list that can be used for agreed upon
 *         communication between the service implementation and a custom user
 *         interface.
 *
 * OUTPUT PARAMETERS
 *
 * outdata Free form attribute list.
 *
 *         The service inplementation needs to pass the task information in
 *         the azn_admin_svc_task attribute (using the azn_attrlist_add_entry()
 *         API). Each specific task information will be a value of this
 *         attribute.

 *         The service implementation needs to pass the result strings
 *         information in the azn_admin_svc_results attribute (using the
 *         azn_attrlist_add_entry() API). Typically, these strings would
 *         include error, warning, and information messages.
 *
 *         The service implementation and the custom user interface can
 *         exchange information of  their choice by setting other attributes
 *         (using the azn_attrlist_add_entry_* APIs).
 *
 * RETURN VALUE
 *      AZN_S_COMPLETE on success, error code on failure
 */
    AZN_DECLSPEC
    azn_status_t
    AZN_CALLTYPE
    azn_admin_get_tasklist(
                          azn_creds_h_t       creds,
                          azn_string_t        locale,
                          azn_attrlist_h_t    indata,
                          azn_attrlist_h_t    outdata)
    {
        printf("%s -- azn_admin_get_tasklist() - ENTRY\n", ADMIN_PLUGIN_VERSION);

        printf("locale: %s\n", locale);

        /*
         * Code needs to be added below to check the return code for all
         * the aznAPI calls
         */

        azn_attrlist_add_entry(outdata, azn_admin_svc_task, "task1");
        azn_attrlist_add_entry(outdata, azn_admin_svc_task, "task2");
        azn_attrlist_add_entry(outdata, azn_admin_svc_task, "task3");
        azn_attrlist_add_entry(outdata, azn_admin_svc_results, "result1");
        azn_attrlist_add_entry(outdata, azn_admin_svc_results, "result2");
        azn_attrlist_add_entry(outdata, azn_admin_svc_results, "result3");

        printf("%s -- azn_admin_get_tasklist() - EXIT\n", ADMIN_PLUGIN_VERSION);


        return(azn_util_errcode(AZN_S_COMPLETE, 0));
    }

/* Make sure this file can be compiled using either the C or C++ compiler */
#ifdef __cplusplus
}
#endif

